/*
 * HUZZAH ESP8266 MQTT client for EVSE
 * Publish OpenEVSE settings and data to an MQTT broker.
 *
 * Main sketch.
 * Code to access the EVSE RAPI was derived from:
 * 'OpenEVSE_RAPI_WiFi_ESP8266.ino'
 * 'OpenEVSE/rapi_proc.cpp'
 *
 * The MQTT client is in 'mqtt_pub.cpp'.
 * MQTT broker access is defined in 'mqtt_pub.h'.
 * WiFI access is defined in 'net_connect.h', 
 */

/*

The MIT License (MIT)

Copyright (c) 2016 Greg Mansfield

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/


#include "mqtt_pub.h"

#define CLIENT_NAME "eGolfEVSE"
#define TOPIC_BASE "eGolf/evse"

/* From J1772EvseController.h */
#define EVSE_STATE_UNKNOWN 0x00
#define EVSE_STATE_A       0x01 // vehicle state A 12V - not connected
#define EVSE_STATE_B       0x02 // vehicle state B 9V - connected, ready
#define EVSE_STATE_C       0x03 // vehicle state C 6V - charging
#define EVSE_STATE_D       0x04 // vehicle state D 3V - vent required

/* Data types. */
#define VTYPE_INT 1
#define VTYPE_INT32 2
#define VTYPE_FLOAT 3
#define VTYPE_STR 4
#define VTYPE_HEX 5
#define VTYPE_STATE 6
#define VTYPE_TIM_DUR 7

struct data_list {
	const char *topic_name;
	int rep_flags;
	const char *cmd;
	int val_type;
	void *value;
};
	
/* Private data. */
/* Data from EVSE: prev_value, new_value, difference */
static float amp[3];
static int32_t volt[3];
static float temp1[3] = { 0, 0, 1.0 };
static float temp2[3];
static float temp3[3];
static float session_wh[3];
static float kwh_acc[3];
static int charge_limit[3];
static int time_limit[3];
static int amp_setting[3];
static unsigned int flags[3];
static int cur_scale[3];
static int cur_offset[3];
static unsigned int gfi_trip_cnt[3];
static unsigned int no_gnd_trip_cnt[3];
static unsigned int stuck_relay_trip_cnt[3];
static int min_amps[3];
static int max_amps[3];
static int volt_scale[3];
static int32_t volt_offset[3];
static int amb_thresh[3];
static int ir_thresh[3];
static int state[3];
static int32_t charge_time[3];
static char version[60];
static char rtc[40];

static struct data_list data_list[] = {
	{ "version", 0x10, "$GV*C1", VTYPE_STR, &version }, // ^C1
	{ "rtc", 0x10, "$GT*BF", VTYPE_STR, &rtc }, // ^BF
	{ "time_limit", 0x10, "$G3^50", VTYPE_INT, &time_limit },
	{ "cur_scale", 0x10, "$GA*AC", VTYPE_INT, &cur_scale }, // ^22
	{ "cur_offset", 0x10, NULL, VTYPE_INT, &cur_offset },
	{ "min_amps", 0x10, "$GC*AE", VTYPE_INT, &min_amps }, // ^20
	{ "max_amps", 0x10, NULL, VTYPE_INT, &max_amps },
	{ "amp_setting", 0x10, "$GE*B0", VTYPE_INT, &amp_setting }, // ^26
	{ "flags", 0x00, NULL, VTYPE_HEX, &flags },
	{ "gfi_trip_cnt", 0x00, "$GF*B1", VTYPE_INT, &gfi_trip_cnt }, // ^25
	{ "no_gnd_trip_cnt", 0x00, NULL, VTYPE_INT, &no_gnd_trip_cnt },
	{ "stuck_relay_trip_cnt", 0x00, NULL, VTYPE_INT, &stuck_relay_trip_cnt },
	{ "voltage", 0x00, "$GG*B2", VTYPE_INT32, &volt }, // ^ 24
	{ "current", 0x00, NULL, VTYPE_FLOAT, &amp },
	{ "charge_limit", 0x10, "$GH^2B", VTYPE_INT, &charge_limit },
	{ "volt_scale", 0x10, "$GM^2E", VTYPE_INT, &volt_scale },
	{ "volt_offset", 0x10, NULL, VTYPE_INT, &volt_offset },
	{ "amb_thresh", 0x10, "$GO^2C", VTYPE_INT, &amb_thresh },
	{ "ir_thresh", 0x10, NULL, VTYPE_INT, &ir_thresh },
	{ "temp1", 0x00, "$GP*BB", VTYPE_FLOAT, &temp1 }, // ^33
	{ "temp2", 0x00, NULL, VTYPE_FLOAT, &temp2 },
	{ "temp3", 0x00, NULL, VTYPE_FLOAT, &temp3 },
	{ "state", 0x00, "$GS*BE", VTYPE_STATE, &state }, // ^BE
	{ "charge_time", 0x00, NULL, VTYPE_TIM_DUR, &charge_time },
	{ "session_whr", 0x00, "$GU*C0", VTYPE_FLOAT, &session_wh }, // ^36
	{ "total_kWh", 0x00, NULL, VTYPE_FLOAT, &kwh_acc },
	{ NULL }
};

static struct data_list *dl;
static unsigned long last_time;
static boolean first;

/* Private functions. */
static void process_line(char par, String line);

void
setup()
{
	Serial.begin(115200);
	last_time = millis() - publish_interval;
	first = true;
	/* Set previous value to max. */
	for (dl = &data_list[0]; dl->topic_name != NULL; dl++) {
		switch (dl->val_type) {

		case VTYPE_HEX:
		case VTYPE_TIM_DUR:
		case VTYPE_STATE:
		case VTYPE_INT: {
			int *value;

			value = (int *)dl->value;
			value[0] = 32767;
			if (value[2] == 0)
				value[2] = 1;
			} break;

		case VTYPE_INT32: {
			int32_t *value;

			value = (int32_t *)dl->value;
			if (value[2] == 0)
				value[2] = 1;
			value[0] = 2147483647L;
			} break;

		case VTYPE_FLOAT: {
			float *value;

			value = (float *)dl->value;
			value[0] = 3.4028235E+38;
			if (value[2] == 0)
				value[2] = 0.5;
			} break;
		}				
	}
	mqtt_setup(CLIENT_NAME, TOPIC_BASE);
}

void
loop()
{
	unsigned long now;
	String line;

	mqtt_loop();

	now = millis();
	if (now - last_time > publish_interval) {
		last_time = now;
	
		/* Get data from EVSE. */
		for (dl = &data_list[0]; dl->topic_name != NULL; dl++) {
			if (!first && (dl->rep_flags & 0x10))
				continue;
			if (dl->cmd == NULL)
				continue;
			Serial.flush();
			Serial.println(dl->cmd);
			delay(100);
			while (Serial.available()) {
				line = Serial.readStringUntil('\r');
				if (line.startsWith("$OK "))
					process_line(dl->cmd[2], line);
			}
			delay(100);
			
		}
	
		/* Publish data. */
		for (dl = &data_list[0]; dl->topic_name != NULL; dl++) {
			char buf[32];

			if (!first && (dl->rep_flags & 0x10))
				continue;
			switch (dl->val_type) {

			case VTYPE_INT: {
				int *value;

				value = (int *)dl->value;
				if (abs(value[1] - value[0]) < value[2])
					break;
				value[0] = value[1];
				mqtt_publish(dl->topic_name,
					     String(value[0]).c_str());
				} break;

			case VTYPE_INT32: {
				int32_t *value;

				value = (int32_t *)dl->value;
				if (abs(value[1] - value[0]) < value[2])
					break;
				value[0] = value[1];
				mqtt_publish(dl->topic_name,
					     String(value[0]).c_str());
				} break;

			case VTYPE_FLOAT: {
				float *value;

				value = (float *)dl->value;
				if (abs(value[1] - value[0]) < value[2])
					break;
				value[0] = value[1];
				mqtt_publish(dl->topic_name,
					     String(value[0]).c_str());
				} break;

			case VTYPE_STR: {
				mqtt_publish(dl->topic_name,
					     (char *)dl->value);
				} break;

			case VTYPE_HEX: {
				int *value;

				value = (int *)dl->value;
				if (abs(value[1] - value[0]) < value[2])
					break;
				value[0] = value[1];
				snprintf(buf, sizeof(buf), "0x%04x", value[0]);
				mqtt_publish(dl->topic_name, buf);
				} break;

			case VTYPE_STATE: {
				int *value;
				const char *state;

				value = (int *)dl->value;
				if (abs(value[1] - value[0]) < value[2])
					break;
				value[0] = value[1];
				mqtt_publish(dl->topic_name,
					     String(value[0]).c_str());
				switch (value[0]) {
				case EVSE_STATE_A:
					state = "NC";
					break;
				case EVSE_STATE_B:
					state = "Ready";
					break;
				case EVSE_STATE_C:
					state = "Charge";
					break;
				case EVSE_STATE_D:
					state = "Vent";
					break;
				default:
					state = NULL;
					break;
				}
				if (state != NULL)
					mqtt_publish("state_s", state);
				} break;

			case VTYPE_TIM_DUR: {
				int *value;
				int h, m, sec, rem;

				value = (int *)dl->value;
				if (abs(value[1] - value[0]) < value[2])
					break;
				value[0] = value[1];
				h = value[0] / 3600;
				rem = value[0] % 3600;
				m = rem / 60;
				sec = rem % 60;
				if (h != 0)
					snprintf(buf, sizeof(buf),
						 "%d:%02d:%02d", h, m, sec);
				else if (m != 0)
					snprintf(buf, sizeof(buf), "%d:%02d",
						 m, sec);
				else
					snprintf(buf, sizeof(buf), "%d", sec);
				mqtt_publish(dl->topic_name, buf);
				} break;
				
			}
		}
		first = false;
	}
}

static void
process_line(char par, String line)
{
	String s;
	int i;

	i = line.indexOf(' ') + 1;
	s = line.substring(i);
	switch (par) {

	case '3': // Get time limit
		time_limit[1] = s.toInt();
		break;

	case 'A':	// Get ammeter settings
		cur_scale[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		cur_offset[1] = s.toInt();
		break;

	case 'C':	// Get current capacity range in amps
		min_amps[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		max_amps[1] = s.toInt();
		break;
		
	case 'E':	// Get settings
		amp_setting[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		flags[1] = s.toInt();
		break;

	case 'F':	// get fault counters
		gfi_trip_cnt[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		no_gnd_trip_cnt[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		stuck_relay_trip_cnt[1] = s.toInt();
		break;

	case 'G':	// get charging current and voltage
		amp[1] = s.toInt() / 1000.;
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		volt[1] = s.toInt();
		break;

	case 'H':	// get charge limit
		charge_limit[1] = s.toInt();
		break;

	case 'M':	// get voltMeter settings
		volt_scale[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		volt_offset[1] = s.toInt();
		break;

	case 'O':	// get Overtemperature thresholds
		amb_thresh[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		ir_thresh[1] = s.toInt();
		break;
		
	case 'P':	// get temPerature
		temp1[1] = s.toInt() / 10.;
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		temp2[1] = s.toInt() / 10.;
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		temp3[1] = s.toInt() / 10.;
		break;

	case 'S':	// get state
		state[1] = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		charge_time[1] = s.toInt();
		break;
		
	case 'T': {	// get time (RTC)
		/* "15 2 19 16 58 53" */
		int y, mo, d, h, m, sc;

		y = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		mo = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		d = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		h = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		m = s.toInt();
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		sc = s.toInt();
		snprintf(rtc, sizeof(rtc), "%d-%02d-%02dT%02d:%02d:%02d",
			 y + 2000, mo, d, h, m, sc);
		} break;

	case 'U':	// get usage
		session_wh[1] = s.toInt() / 3600.;
		i = s.indexOf(' ') + 1;
		s = s.substring(i);
		kwh_acc[1] = s.toInt() / 1000;
		break;	

	case 'V':       // get version
		s.toCharArray(version, sizeof(version)-1);
		break;
		
	default:
		break;
	}
}		
